#define _USE_MATH_DEFINES
#include <cmath>
#include <utility>
#include <tuple>
#include <random>
#include <algorithm>

#include "ofApp.h"


namespace {

std::pair<float, float> center() {
    return{ ofGetViewportWidth() / 2.f, ofGetViewportHeight() / 2.f };
}

const float paddle_width = 10;
const float paddle_height = 100;
const float paddle_speed = 10;
const float ball_radius = 5;
const float ball_base_speed = 10;
const float ball_paddle_speedup = ball_base_speed / 10;
const float max_bounce_angle = 5 * M_PI / 12;

auto& get_rng() {
    static std::mt19937_64 mt{ std::random_device{}() };
    return mt;
}

std::pair<float, float> random_vector() {
    auto& rng = get_rng();
    static std::uniform_real_distribution<float> xdist(0.5, 0.8);
    static std::uniform_real_distribution<float> ydist(-0.5, 0.5);
    bool halfBool = ydist(rng) < 0;
    return{ halfBool? xdist(rng) : -xdist(rng),
                   ydist(rng) };
}

}




//--------------------------------------------------------------
void ofApp::setup(){
    bg_color = ofColor::black;
    ofSetBackgroundColor(bg_color);
    resetBall();
    stopBall();
    ofSetFrameRate(target_fps);
    frame_rate_multip = 1.;
    p1Score = p2Score = 0;
    scoreFont.load(OF_TTF_MONO, 32);
}

void ofApp::handleKeys() {
    for (auto key : keySet) {
        switch (key) {
        case OF_KEY_UP:
            yspeed += -1;
            break;
        case OF_KEY_DOWN:
            yspeed += 1;
            break;
        case OF_KEY_LEFT:
            xspeed += -1;
            break;
        case OF_KEY_RIGHT:
            xspeed += 1;
            break;
        case 's':
            stopBall();
            break;
        case 'c':
            resetBall();
            break;
        case 'o':
            right_paddle_pos += -paddle_speed * frame_rate_multip;
            break;
        case 'l':
            right_paddle_pos += paddle_speed * frame_rate_multip;
            break;
        case 'e':
            left_paddle_pos += -paddle_speed * frame_rate_multip;
            break;
        case 'd':
            left_paddle_pos += paddle_speed * frame_rate_multip;
            break;
        }
    }
}

void ofApp::stopBall() {
    xspeed = 0;
    yspeed = 0;
}

void ofApp::resetBall() {
    std::tie(xpos, ypos) = center();
}

void ofApp::startBall() {
    std::tie(xspeed, yspeed) = random_vector();
    xspeed *= ball_base_speed;
    yspeed *= ball_base_speed;
}

float ofApp::fpsAdjusted(float movement) const {
    return movement * frame_rate_multip;
}

void ofApp::checkPaddleCollision() {
    auto xstep = fpsAdjusted(xspeed);
    auto ystep = fpsAdjusted(yspeed);
    // Collision with right-side paddle
    if (xpos + xstep + ball_radius >= ofGetViewportWidth() - paddle_width 
        && ypos + ystep > right_paddle_pos 
        && ypos + ystep < right_paddle_pos + paddle_height) {
        xspeed *= -1;
        xspeed += -ball_paddle_speedup;
        // change y speed according to offset from middle
        auto intersectY = right_paddle_pos + paddle_height / 2 - (ypos + ystep);
        auto normalizedIntersectY = intersectY / (paddle_height / 2);
        auto bounce_angle = -normalizedIntersectY * max_bounce_angle;
        yspeed = xspeed * -sin(bounce_angle);
    }
    
    // Collision with left-side paddle
    if (xpos + xstep + ball_radius < 0 + paddle_width
        && ypos + ystep > left_paddle_pos
        && ypos + ystep < left_paddle_pos + paddle_height) {
        xspeed *= -1;
        xspeed += ball_paddle_speedup;
        // change y speed according to offset from middle
        auto intersectY = left_paddle_pos + paddle_height / 2 - (ypos + ystep);
        auto normalizedIntersectY = intersectY / (paddle_height / 2);
        auto bounce_angle = normalizedIntersectY * max_bounce_angle;
        yspeed = xspeed * -sin(bounce_angle);
    }
}

void ofApp::checkWallCollision() {
    auto xstep = fpsAdjusted(xspeed);
    auto ystep = fpsAdjusted(yspeed);
    // on x axis, it should count a score and reset.
    if (xpos + xstep + ball_radius >= ofGetViewportWidth() || xpos + xstep < 0) {
        if (xpos + xstep + ball_radius >= ofGetViewportWidth()) {
            p1Score++;
        } else {
            p2Score++;
        }
        stopBall();
        resetBall();
        startBall();
    }
    // y axis is just bounce
    if (ypos + ystep + ball_radius >= ofGetViewportHeight() || ypos + ystep < 0) {
        yspeed *= -1;
    }
    xpos += fpsAdjusted(xspeed);
    ypos += fpsAdjusted(yspeed);
}

//--------------------------------------------------------------
void ofApp::update(){
    // Adjust updates by frame rate first
    frame_rate_multip = target_fps / (1. / ofGetLastFrameTime());

    // Then handle keys
    handleKeys();

    // Finally collisions
    checkPaddleCollision();
    checkWallCollision();
}

//--------------------------------------------------------------
void ofApp::draw(){
    ofSetColor(ofColor::red);
    ofCircle(xpos, ypos, 5);
    ofPoint rectloc1 = ofPoint(0, left_paddle_pos);
    ofPoint rectloc2 = ofPoint(ofGetViewportWidth() - paddle_width, right_paddle_pos);
    ofRect(rectloc1, paddle_width, paddle_height);
    ofRect(rectloc2, paddle_width, paddle_height);
    std::stringstream strm;
    strm << "fps: " << ofGetFrameRate();
    ofSetWindowTitle(strm.str());
    ofSetColor(ofColor::white);
    // Properly center current score
    auto target = ofGetViewportWidth() / 2;
    auto result_string = std::to_string(p1Score) + " : " + std::to_string(p2Score);
    auto width = scoreFont.stringWidth(result_string);
    target -= width / 2;
    scoreFont.drawString(result_string, target, 100);
}

//--------------------------------------------------------------
void ofApp::keyPressed(int key){
    keySet.insert(key);
}

//--------------------------------------------------------------
void ofApp::keyReleased(int key){
    keySet.erase(key);
}

//--------------------------------------------------------------
void ofApp::mouseMoved(int x, int y ){

}

//--------------------------------------------------------------
void ofApp::mouseDragged(int x, int y, int button){

}

//--------------------------------------------------------------
void ofApp::mousePressed(int x, int y, int button){

}

//--------------------------------------------------------------
void ofApp::mouseReleased(int x, int y, int button){

}

//--------------------------------------------------------------
void ofApp::mouseEntered(int x, int y){

}

//--------------------------------------------------------------
void ofApp::mouseExited(int x, int y){

}

//--------------------------------------------------------------
void ofApp::windowResized(int w, int h){

}

//--------------------------------------------------------------
void ofApp::gotMessage(ofMessage msg){

}

//--------------------------------------------------------------
void ofApp::dragEvent(ofDragInfo dragInfo){ 

}
